/*
 * Interface for accessing iRiver's IFP devices
 * $Id: ifp.h,v 1.21 2004/10/04 14:23:17 oakhamg Exp $
 *
 * Copyright (C) Geoff Oakham, 2004; <oakhamg@users.sourceforge.net>
 *
 * This driver is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free
 * Software Foundation; only version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License in the
 * files "COPYING" or "COPYING.iriverfs"; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef IFP_H
#define IFP_H 1

#include "ifp_os.h"

//These can all be enabled from the "make menuconfig" interface now
//#define IFP_DEBUG_SYSCALL 1
//#define IFP_DEBUG_USB_SNOOPING	1
//#define IFP_DEBUG_USB_RAWDATA	1

#define IFP_BULK_MAXPATHLEN	0x400 //ie, utf16
#define IFP_MAXPATHLEN		0x200 //ie, utf8
#define IFP_BUFFER_SIZE		IFP_BULK_MAXPATHLEN

#define IFP_BULK_BUFF_BITS 14
#define IFP_BULK_BUFF_SIZE 0x4000

//Members of this structure are considered private (ie for internal use only).
//A future version may replace this definition with a forward declaration.
struct ifp_device {
	int bulk_to;
	int bulk_from;
	int model; //decimal: eg 100, 300, 500, 700, 800
	//do we need access to the firmware version as well?
	
	//These buffers are use for servicing individual function calls.  Contents
	//are not preserved between calls.
	
	//b1 is used by:
	//	push_unicode
	//	pop_unicode
	//	ifp_dir_next_debug
	//	ifp_set_fat_page
	ifp_u8 b1[IFP_BUFFER_SIZE];

	//b2 is used by:
	//	swap_fat_entries	(ifp_rename* helper)
	//	_ifp_list_dirs_debug	(ifp_rename* helper)
	//	_ifp_list_dirs		(ifp_list_dirs helper)
	//	ifp_device_info
	ifp_u8 b2[IFP_BUFFER_SIZE];

	//b3 is used by:
	//	swap_fat_entries	(ifp_rename* helper)
	//	swap_filenames		(ifp_rename* helper)
	//	get_file_size		(_ifp_list_dirs helper)
	ifp_u8 b3[IFP_BUFFER_SIZE];

	//The following data is state that persists between library calls.
	int last_buffer_size;

	#define IFP_MODE_NONE (0)
	#define IFP_MODE_READING (1)
	#define IFP_MODE_WRITING (2)

	//data buffer.  Contains the last block read, or part of the block
	//being written.  The current position (for reading or writing) is
	//current_offset % IFP_BULK_BUFF_SIZE.
	ifp_u8 iobuff[IFP_BULK_BUFF_SIZE];

	int mode;
	loff_t current_offset;
	loff_t filesize;
	ifp_u8 dirname[IFP_BUFFER_SIZE];
	ifp_u8 filename[IFP_BUFFER_SIZE];

	//sanity counters
	int readcount;
	int alt_readcount;

	//usb device handle (system-specific)
	void * private;
};

#define IFP_FILE                1
#define IFP_DIR                 2

int ifp_init(struct ifp_device * dev, void * dev_handle);
int ifp_finalize(struct ifp_device * dev);

int ifp_selftest(struct ifp_device * dev);

//status and misc
int ifp_device_info(struct ifp_device * dev, char * s, int n);
int ifp_battery(struct ifp_device * dev);
int ifp_capacity(struct ifp_device * dev);
int ifp_freespace(struct ifp_device * dev); //better name?
int ifp_model(struct ifp_device * dev, char * s, int n);
int ifp_delta(struct ifp_device * dev, int * values);  //int values[4]
//returns a raw bcd value.  For human consumption, I suggest ("%x.%02x", r/0x100, r%0x100).
int ifp_firmware_version(struct ifp_device * dev);

// reading files
int ifp_read_open(struct ifp_device * dev, char * f, int n);
int ifp_read_close(struct ifp_device * dev);
int ifp_read_seek(struct ifp_device * dev, int abs_position);
	//returns the actual number of bytes read
int ifp_read_data(struct ifp_device * dev, void * buff, int bytes);
int ifp_read_eof(struct ifp_device * dev);

// writing files
int ifp_write_open(struct ifp_device * dev, char * f, int n, int fsize);
int ifp_write_close(struct ifp_device * dev);
int ifp_write_data(struct ifp_device * dev, void * buff, int bytes);

// rename
// API TOUGHT: would it be better for the library to detect if the
// remote thing is a file/dir and call the appropriate method?  The
// advantage to this solution, is we don't have to issue a test if
// the caller knows ahead of time what they are doing.
//
// 	-decide if this works for files and directories or just files.
int ifp_rename_file(struct ifp_device * dev,
	char * old_file, char * new_file, int maxlength);
int ifp_rename_dir(struct ifp_device * dev,
	char * old_file, char * new_file, int maxlength);

// metadata (file & directory) manipulation
int ifp_delete(struct ifp_device * dev, char * f, int n);
int ifp_mkdir(struct ifp_device * dev, char * f, int n);
int ifp_rmdir(struct ifp_device * dev, char * f, int n);

//The callback should return 0 on succes or 1 to stop without error.
//(Any other value will be interpreted as an error.)  The parameters are:
//	int callbk(void * context, int type, char * name, int n, int filesize)
//'type' is either IFP_FILE or IFP_DIR
//'filesize' is undefined for directories,
//'name' is the file or directory name *without* its path
//'context' can be NULL.
//
//API THOUGHT: consider passing in the full path
int ifp_list_dirs(struct ifp_device * dev, char * dirname, int n,
	int(*callbk)(void *, int, char *, int, int),
	void * context);

#ifdef IFP_FUTURE
//boolean tests
int ifp_is_file(struct ifp_device * dev, char * f, int n);
int ifp_is_dir (struct ifp_device * dev, char * f, int n);
//returns IFP_FILE, IFP_DIR, 0 (none), or <0 (error)
int ifp_exists (struct ifp_device * dev, char * f, int n);

int ifp_file_size(struct ifp_device * dev, char * f, int n);

int ifp_read_filesize(struct ifp_device * dev);
int ifp_read_pos(struct ifp_device * dev);

//missing functionality:
int ifp_format(struct ifp_device * dev);
//int ifp_update_firmware(struct ifp_device * dev);

#ifndef __KERNEL__
//userland-only convience functions
int ifp_read_file(struct ifp_device * dev, FILE * dst, char * path, int n);
int ifp_write_file(struct ifp_device * dev, char * path, int n, FILE * src);

//for user feedback
int ifp_read_file_progress(struct ifp_device * dev, FILE * dst, char * path,
	int n,      int(*progress)(void *, int), void * context);
int ifp_write_file_progress(struct ifp_device * dev, char * path, int n,
	FILE * src, int(*progress)(void *, int), void * context);
#endif
#endif
#endif // IFP_H

