/*
 * ifp supporting functions.. "compound" ones
 * $Id: ifp_comp.c,v 1.16 2004/10/04 14:23:17 oakhamg Exp $
 *
 * Copyright (C) Geoff Oakham, 2004; <oakhamg@users.sourceforge.net>
 */

#include "ifp.h"
#include "ifp_os.h"
#include "ifp_prim.h"

//quick hack
#define IFP_ENOENT 2
#define IFP_EIO    5

static char * battery_status_art [] = {
    "=[    ]",
    "=[#   ]",
    "=[##  ]",
    "=[### ]",
    "=[####]",
};

int ifp_device_info(struct ifp_device * dev, char * s, int n) {
    char * p = dev->b2;
    char * battery_info = NULL;
    int i = 0;
    int version, battery;
    int deltas[4];

    i = ifp_model(dev, p, IFP_BUFFER_SIZE);
    if (i) {
        ifp_err_i(i, "error getting model number.");
        p[0] = '-';
        p[1] = '\0';
    }

    version = ifp_firmware_version(dev);
    if (version < 0) {
        ifp_err_i(version, "error getting firmware version.");
        i = version;
    }

    battery = ifp_battery(dev);
    if (battery < 0) {
        ifp_err_i(battery, "error getting firmware version.");
        battery_info = "=[fubr]";
        i = battery;
    } else if (battery > 4) {
        battery_info = "=[????]";
    } else {
        battery_info = battery_status_art[battery];
    }

    i = ifp_delta(dev, deltas);
    if (i) {
        ifp_wrn("problem getting delta value (error code %d).", i);
        //experimental feature, don't return the error
        i = 0;
    }

    snprintf(s, n, "model %s, firmware %x.%02x, battery %s"
            ", delta %d.%d.%d.%d"
            , p, version/0x100, version % 0x100, battery_info
            , deltas[0], deltas[1], deltas[2], deltas[3]
            );

    return i;
}

static int noop_dir_callbk(void * p, int t, char * s, int n, int fs)
{
    int * pn = p;
    if (*pn == 5) {
        //return early--save time.
        return 1;
    }
    (*pn)++;
    return 0;
}

int ifp_selftest(struct ifp_device * dev) {
    int i = 0;
    int n = 0;

    i = ifp_firmware_version(dev);
    if (i < 0) {
        ifp_err_i(i, "couldn't get firmware version.");
        return i;
    }

    i = ifp_battery(dev);
    if (i < 0) {
        ifp_err_i(i, "couldn't get battery status.");
        return i;
    }

    i = ifp_list_dirs(dev, "\\", 1, noop_dir_callbk, &n);
    if (i) {
        ifp_err_i(i, "couldn't get basic directory listing.");
        return i;
    }

    if (n < 3) {
        ifp_wrn("only %d items could be found in the root directory.  Either there's a problem, or the device is empty.", n);
    }

    return i;
}

static int touch(struct ifp_device * dev, char * dir, char * file, int n)
{
	int i = 0;

	//ifp_dbg("[ifp touch] touching %s\n", (char *)file);
	//ifp_dbg("[ifp touch] in %s\n", (char *)dir);

	i = ifp_dir_open(dev, dir, n);
        ifp_err_jump(i, out, "couldn't open dir.");
	i = ifp_file_open_new(dev, file, n, 0);
        ifp_err_jump(i, out, "couldn't create file.");
	i = ifp_file_close(dev);
        ifp_err_jump(i, out, "couldn't close file.");
	i = ifp_dir_close(dev);
        ifp_err_jump(i, out, "couldn't close dir.");

out:
	return i;
}

int _ifp_list_dirs_debug(
	struct ifp_device * dev,
	char * name, int n,
	int(*callbk)(void *, int, char *, int, int, int, int),
	void * context)
{
	//unsigned char utf8_buf[IFP_BULK_MAXPATHLEN];
	ifp_u8 * utf8_buf = dev->b2;
	int i = 0;

	i = ifp_dir_open(dev, name, n);
	ifp_err_jump(i, out, "dir.open failed.");

	while(1) {
		int dir, page, offset, chipmunk;
		int ret;
		ret = ifp_dir_next_debug(dev, utf8_buf, IFP_BUFFER_SIZE,
			IFP_FILE|IFP_DIR, &dir, &page, &offset, &chipmunk);
		if (ret < 0) {
			ifp_err_i(ret, "error sending dir.next.");
			i = ret; goto out;
		} else if (ret == 0) {
			//ifp_dbg("Done.");
			break;
		} else {
			i = callbk(context, ret, utf8_buf,
				strnlen(utf8_buf, IFP_BUFFER_SIZE),
				dir, page, offset);
			if (i < 0) {
				ifp_err_i(i, "callback returned error.");
				break;
			} else if (i == 1) {
				//ifp_dbg("callback requested early break.");
				break;
			} else if (i != 0) {
				ifp_err_i(i, "callback returned a wierd value.");
				i = -1;
				goto out;
			} else {
				//ifp_dbg("callback returned %d for '%s'.",
				//	i, utf8_buf);
			}
		}
	}

	i = ifp_dir_close(dev);
	ifp_err_jump(i, out, "dir.close failed.");

	i = 0;
out:
	return i;
}

static int get_id_callback(void * context, int type, char * name, int n,
	int dir, int page, int offset)
{
	void ** pp = (void **) context;
	char * target = pp[0];
	int * target_width = pp[1];
	int * pdir = pp[2];
	int * ppage = pp[3];
	int * poffset = pp[4];
	int * found = pp[5];

	//ifp_dbg("dir=%#x page=%#x, offset=%#x, name='%s'", dir, page, offset, name);
	if ( *target_width == n) {
		if (strncmp(target, name, n)==0)
		{
			//printk("[get_id_callback] '%s' matches '%s'.\n",
			//	target, name);
			*pdir = dir;
			*ppage = page;
			*poffset = offset;
			*found = 1;
			return 1;
		}
	}

	return 0;
}

static int get_file_id(struct ifp_device * dev, char * dir, char * file, int n,
	int * id_dir, int * id_page, int * id_off)
{
	int i = 0;
	int filelen;
	int dirlen;
	int found = 0;
	void ** context[6] = {(void*)file, (void*)&filelen, (void*)id_dir,
		(void*)id_page, (void*)id_off, (void*)&found};
	dirlen = strnlen(dir, n);
	//one extra for the path separator
	filelen = strnlen(file, n) - dirlen;
	file += dirlen;
	if (dirlen > 1) {
		filelen--;
		file++;
	}
	//ifp_dbg("file='%s', filelen=%d, dir='%s', dirlen=%d", file, filelen,
	//	dir, dirlen);
	if (filelen < 0) {
		ifp_err_i(i, "bailing");
		return -1;
	}
	IFP_BUG_ON(filelen < 0);
	context[0] = (void *)(file);

	//FIXME: return an error on not finding a matching entry.
	i = _ifp_list_dirs_debug(dev, dir, n, get_id_callback, (void *)context);
	ifp_err_jump(i, out, "list failed.");
	if (found == 0) {
		ifp_err("entry not found for dir='%s' file='%s'", dir, file);
		i = -1;
	}
	//ifp_dbg("returning dir=%#x, page=%#x, off=%#x", *id_dir, *id_page, *id_off);

out:
	return i;
}

static inline int calc_fat_address(
	int base_page,
	int slot_number,
	int record_offset,
	int * page,
	int * offset
	)
{
	const int page_size  = 0x0200;
	const int slot_width = 0x0020;
	IFP_BUG_ON(page == NULL);
	IFP_BUG_ON(offset == NULL);

	*page = base_page;
	*offset = slot_width * slot_number + record_offset;

	if (*offset >= page_size) {
		(*page) += *offset / page_size;
		(*offset) %= page_size;
	}
	return 0;
}


static int swap_fat_entries(struct ifp_device * dev,
	int id_dirA, int id_pageA, int id_offA,
	int id_dirB, int id_pageB, int id_offB)
{
	const int data_ptr_size = 8;
	const int page_size  = 0x0200;
	ifp_u8 tmp[data_ptr_size];

	ifp_u8 * _buffA = dev->b2, * _buffB = dev->b3;
	ifp_u8 * buffA = _buffA, * buffB = _buffB;

	int i = 0;

	i = calc_fat_address(id_pageA, id_offA, 0x040 - data_ptr_size,
		&id_pageA, &id_offA);
	ifp_err_jump(i, out, "cal failed. (a)");

	i = calc_fat_address(id_pageB, id_offB, 0x040 - data_ptr_size,
		&id_pageB, &id_offB);
	ifp_err_jump(i, out, "cal failed. (b)");

	i = ifp_get_fat_page(dev, id_dirA, id_pageA, _buffA, page_size);
	ifp_err_jump(i, out, "read failed for page=%#x, dir=%#x (a)",
		id_pageA, id_dirA);

	if (id_dirA == id_dirB && id_pageA == id_pageB) {
		//ifp_dbg("same page number, using same buffers.");
		buffB = _buffA;
	} else {
		i = ifp_get_fat_page(dev, id_dirB, id_pageB, _buffB, page_size);
		ifp_err_jump(i, out, "read failed for page=%#x, dir=%#x (b)",
			id_pageB, id_dirB);
	}

	buffA += id_offA;
	buffB += id_offB;

	memcpy(tmp,   buffA, data_ptr_size);
	memcpy(buffA, buffB, data_ptr_size);
	memcpy(buffB, tmp,   data_ptr_size);

	i = ifp_set_fat_page(dev, id_dirA, id_pageA, _buffA, page_size);
	ifp_err_jump(i, out, "write failed. (a)");

	if (id_dirA != id_dirB || id_pageA != id_pageB) {
		i = ifp_set_fat_page(dev, id_dirB, id_pageB, _buffB, page_size);
		ifp_err_jump(i, out, "write failed. (b)");
	}

out:

	return i;
}

static int swap_filenames(struct ifp_device * dev,
	char * oldpath, char * newpath, int max_string)
{
	int id_dirA, id_dirB, id_pageA, id_pageB, id_offA, id_offB;
	int i;
	char * b = dev->b3;

	i = ifp_copy_parent_string(b, oldpath, max_string);
	ifp_err_jump(i, out, "getting parent directory of %s failed.",oldpath);
	i = get_file_id(dev, b, oldpath, max_string,
		&id_dirA, &id_pageA, &id_offA);
	ifp_err_jump(i, out, "file_id failed. (a)");

	i = ifp_copy_parent_string(b, newpath, max_string);
	ifp_err_jump(i, out, "getting parent directory of %s failed.",newpath);
	i = get_file_id(dev, b, newpath, max_string,
		&id_dirB, &id_pageB, &id_offB);
	ifp_err_jump(i, out, "file_id failed. (b)");

	//uses b1,b2,b3.
	i = swap_fat_entries(dev, id_dirA, id_pageA, id_offA,
				id_dirB, id_pageB, id_offB);
	ifp_err_jump(i, out, "swap failed. "
		"dirA=%#x, pageA=%#x, offA=%#x, dirB=%#x, pageB=%#x, offB=%#x",
		id_dirA, id_pageA, id_offA, id_dirB, id_pageB, id_offB);

out:
	return i;
}

int ifp_rename_file(struct ifp_device * dev,
	char * old_file, char * new_file, int max_string)
{
	int i;
	char * b = dev->b3;

	i = ifp_copy_parent_string(b, new_file, max_string);
	ifp_err_jump(i, out, "getting parent directory of %s failed.",new_file);
	i = touch(dev, b, new_file, max_string);
	ifp_err_jump(i, out, "Touch failed.");

	i = swap_filenames(dev, old_file, new_file, max_string);
	ifp_err_jump(i, out, "filename swap failed.");

	i = ifp_delete(dev, old_file, max_string);
	ifp_err_jump(i, out, "delete failed.");

out:
	return i;
}

int ifp_rename_dir(struct ifp_device * dev,
	char * old_dir, char * new_dir, int max_string)
{
	int i;

	i = ifp_mkdir(dev, new_dir, max_string);
	ifp_err_jump(i, out, "mkdir failed.");

	i = swap_filenames(dev, old_dir, new_dir, max_string);
	ifp_err_jump(i, out, "filename swap failed.");

	i = ifp_rmdir(dev, old_dir, max_string);
	ifp_err_jump(i, out, "rmdir failed.");

out:
	return i;
}

static int get_file_size(struct ifp_device * dev,
	char * dir, int dsize,
	char * f, int fsize)
{
	int size = 0;
	int i;
	ifp_u8 * buf = dev->b3;
	ifp_u8 * p = buf;
	int dlen, flen;

	dlen = strnlen(dir, dsize);
	flen = strnlen(f, fsize);
	IFP_BUG_ON(dlen + flen + 2 >= IFP_BULK_MAXPATHLEN);
	memcpy(p, dir, dlen); p += dlen;
	if (dlen > 1) {
		//if dir ls simply \ we don't want to make it \\.
		*p = '\\'; p++;
	}
	memcpy(p, f, flen); p += flen;
	*p = 0; p++;


	i = ifp_file_open(dev, buf, IFP_BULK_MAXPATHLEN);
	if (i == 1) {
        	ifp_err("The file '%s' doesn't exist!", buf);
		return -IFP_ENOENT;
	} else if (i) {
        	ifp_err_i(i, "Error opening file '%s'.", buf);
		return i < 0 ? i : -IFP_EIO;
	}

	size = ifp_file_size(dev);
	if (size < 0) {
        	ifp_err_i(size, "Error getting size of '%s'.", buf);
		//fallthrough
	}
	i = ifp_file_close(dev);
	if (i) {
        	ifp_err_i(i, "Error closing file '%s'.", buf);
		return i < 0 ? i : -IFP_EIO;
	}

	return size;
}

int _ifp_list_dirs(
	struct ifp_device * dev,
	char * dirname, int dsize,
	int type,
	int(*callbk)(void *, int, char *, int, int),
	void * context)
{
	ifp_u8 * utf8_buf = dev->b2;
	int i = 0;

	IFP_BUG_ON(sizeof(dev->b2) < IFP_BULK_MAXPATHLEN);

	while(1) {
		int ret;
		ret = ifp_dir_next(dev, utf8_buf, IFP_BULK_MAXPATHLEN, type);
		//IFP_FILE|IFP_DIR)
		if (ret < 0) {
        		ifp_err_i(ret, "Error sending control message dir.next.");
			i = ret;
			return i;
		} else if (ret == 0) {
			//ifp_dbg("Done.");
			break;
		} else {
			int filesize = 0;
			if (ret == IFP_FILE) {
				filesize = get_file_size(dev,
					dirname, dsize, utf8_buf,
					IFP_BULK_MAXPATHLEN);
			}
			if (filesize < 0) {
        			ifp_err_i(filesize, "Error getting filesize.");
				return filesize;
			}

			i = callbk(context, ret, utf8_buf, strlen(utf8_buf),
				filesize);
			if (i < 0) {
        			ifp_err_i(i, "Callback returned error.");
				return i;
			} else if (i == 1) {
				//ifp_dbg("[_ifp_list_dirs] callback requested early break.\n");
				break;
			} else if (i != 0) {
        			ifp_err_i(i, "Callback returned a wierd value.");
				return -IFP_EIO;
			}
		}
	}

	i = 0;
	return i;
}

int ifp_list_dirs(
	struct ifp_device * dev,
	char * filename, int n,
	int(*callbk)(void *, int, char *, int, int),
	void * context)
{
	int i = 0;

	i = ifp_dir_open(dev, filename, n);
	ifp_err_jump(i, out, "dir.open failed.");

	i = _ifp_list_dirs(dev, filename, n, IFP_FILE|IFP_DIR, callbk, context);
	ifp_err_jump(i, out, "_list_dirs failed.");

	i = ifp_dir_close(dev);
	ifp_err_jump(i, out, "dir.close failed.");

out:
	return i;
}

static int _subdir_counter(void * context, int type, char * name, int n, int size)
{
	int * pn = context;
	(*pn) += 1;

	return 0;
}

int ifp_count_subdirs(struct ifp_device * dev, char * dirname, int sn)
{
	int i = 0;
	int n = 0;

	i = ifp_dir_open(dev, dirname, sn);
	if (i) {
		ifp_err_i(i, "dir_open failed");
		return i;
	}

	i = _ifp_list_dirs(dev, dirname, sn, IFP_DIR, _subdir_counter, &n);
	if (i) {
		ifp_err_i(i, "ifp_list_dirs failed");
		return i;
	} else {
		IFP_BUG_ON(n < 0);
	}

	i = ifp_dir_close(dev);
	if (i) {
		ifp_err_i(i, "dir_close failed");
		return i;
	}

	return n;
}


