/*
 * IRiver ifp supporting functions
 * $Id: ifp_readwrite.c,v 1.10 2004/09/30 00:24:27 oakhamg Exp $
 *
 * Copyright (C) Geoff Oakham, 2004; <oakhamg@users.sourceforge.net>
 */

#include "ifp.h"
#include "ifp_os.h"
#include "ifp_prim.h"

#define IFP_EIO  5
#define IFP_ENOSPC 28

//This needs a home.  (Currently declared in ifp_prim.h)
int ifp_copy_parent_string(char * p, char * f, int n) {
	int i = 0;
	char * c;
	c = strrchr(f, '\\');
	if (c == NULL) {
		ifp_err("apparently '%s' doesn't contain a '\\'\n", f);
		i = -1;
		goto out;
	}
	if (c == f) {
		//special case
		p[0] = '\\';
		p[1] = '\0';
	} else {
		memcpy(p, f, (c-f));
		p[c-f] = '\0';
	}

out:
	return i;
}

int ifp_read_open(struct ifp_device * dev, char * f, int n)
{
	int i = 0;
	if (dev->mode != IFP_MODE_NONE) {
		ifp_err("device has been left for %sing.",
			dev->mode == IFP_MODE_READING ? "read" : "writ");
		i = -1;
		goto out;
	}
	IFP_BUG_ON(n > IFP_BUFFER_SIZE);
	memcpy(dev->filename, f, n);
	i = ifp_copy_parent_string(dev->dirname, f, n);
	ifp_err_jump(i, out, "error copying directory name");

	i = ifp_dir_open(dev, dev->dirname, n);
	ifp_err_jump(i, out, "error opening directory '%s'.", dev->dirname);

	i = ifp_file_open(dev, dev->filename, n);
	ifp_err_jump(i, out, "error opening file '%s'.", dev->filename);

	IFP_BUG_ON(n > IFP_BUFFER_SIZE);
	n = n <= IFP_BUFFER_SIZE ? n : IFP_BUFFER_SIZE;
	dev->mode = IFP_MODE_READING;
	dev->current_offset = 0;
	dev->filesize = ifp_file_size(dev);;
	if (dev->filesize < 0) {
		ifp_err_i((int)(dev->filesize), "error reading filesize of '%s'.", f);
		goto out;
	}
	dev->readcount = 0;
	dev->alt_readcount = 0;

	//ifp_dbg("opened %s successfully, returning %d", f, i);

out:
	return i;
}
int ifp_read_close(struct ifp_device * dev)
{
	int i = 0;

	if (dev->alt_readcount != dev->readcount) {
		ifp_err("readcounts don't match.  readcount=%d, alt_readcount=%d",
			dev->readcount, dev->alt_readcount);
	}

	i = ifp_file_close(dev);
	if (i) {
		ifp_err_i(i, "file close failed");
	}
	i = ifp_dir_close(dev);
	if (i) {
		ifp_err_i(i, "dir close failed");
	}
	dev->mode = IFP_MODE_NONE;
	//ifp_dbg("closing %s, returning %d", dev->filename, i);

	return i;
}

//Reads next block.  'bytes' should be IFP_BULK_BUFF_SIZE unless
//it's the last block--in which case 'bytes' is the number of bytes
//we believe are left to download.
static int read_next_block(struct ifp_device * dev, int bytes) {
	int i;
	IFP_BUG_ON(bytes > IFP_BULK_BUFF_SIZE);
	if (dev->readcount * IFP_BULK_BUFF_SIZE + bytes > dev->filesize) {
		ifp_err("Sanity check failed.  We've read %d x 16384 bytes, and"
			" are about to read %d more from a %d byte file.",
			dev->readcount, bytes, (int)dev->filesize);
	}
	i = ifp_file_download(dev, dev->iobuff, IFP_BULK_BUFF_SIZE);
	if (i < 0) {
		ifp_err_i(i, "error reading block at %s+%#lx",
			dev->filename, (long)dev->current_offset);
		goto out;
	} else if (i != bytes) {
		dev->readcount++;
		if (dev->alt_readcount != dev->readcount) {
			ifp_err("readcount=%d, alt_readcount=%d",
				dev->readcount, dev->alt_readcount);
		}
		ifp_err_i(i, "error reading block.. I expected %d bytes but got less; readcount is %d",
			bytes, dev->readcount);
		i = -IFP_EIO;
		goto out;
	} 
	dev->readcount++;
	IFP_BUG_ON(i > bytes);
	i = 0;

	//not strickly necessary..
	if (bytes < IFP_BULK_BUFF_SIZE) {
		memset(dev->iobuff + bytes, 0, IFP_BULK_BUFF_SIZE - bytes);
	}

out:
	return i;
}

//return the number of bytes read.
int ifp_read_data(struct ifp_device * dev, void * buff, int bytes)
{
	int bytes_requested = bytes;
	int bytes_read = 0;
	int n;
	int i = 0;
	int block_off;
	int available;
	ifp_u8 * o = buff;

	bytes = min(bytes, (int)(dev->filesize - dev->current_offset));
	while (bytes > 0) {
		block_off = dev->current_offset % IFP_BULK_BUFF_SIZE;
		if (block_off == 0) {
			//read next block
			available = min(IFP_BULK_BUFF_SIZE,
				(int)(dev->filesize - dev->current_offset));

			i = read_next_block(dev, available);
			ifp_err_jump(i, out, "error reading next block"
				" filesize=%d position=%d", (int)dev->filesize,
				(int)dev->current_offset);
		} else {
			available = IFP_BULK_BUFF_SIZE - block_off;
		}
		n = min(bytes, available);
		IFP_BUG_ON(n <= 0);

		memcpy(o, dev->iobuff + block_off, n);
		o += n;
		dev->current_offset += n;
		bytes -= n;
		bytes_read += n;
	}
	IFP_BUG_ON(bytes < 0);
	IFP_BUG_ON(bytes != 0);

	if (bytes_read != bytes_requested && dev->current_offset != dev->filesize) {
		ifp_dbg("returning %d instead of %d (but not EOF)", bytes_read, bytes_requested);
	}
	IFP_BUG_ON(bytes_read != bytes_requested && dev->current_offset != dev->filesize);

	return bytes_read;
out:
	return i;
}

int ifp_read_eof(struct ifp_device * dev) {
	return dev->current_offset == dev->filesize;
}

static int fake_block_reads(struct ifp_device * dev, int n) {
	int actual, j,i = 0;
	const int blocksize = IFP_BULK_BUFF_SIZE;
	ifp_dbg("here -- sanity check");
	i = _ifp_set_buffer_size(dev, blocksize, 1);
	if (i) {
		ifp_err_i(i, "set buffer failed");
		return i > 0 ? -IFP_EIO : i;
	}
        for (j=0; j < n; j++) {
                i = _ifp_file_download(dev, (int)blocksize, &actual);
		ifp_err_jump(i, out, "download control message failed");
		dev->readcount++;

                if (actual != blocksize) {
                        i = -IFP_EIO;
                        ifp_err("fake read failed,  I can't handle "
                                "getting %d bytes instead of %d\n",
				actual, (int)blocksize);
			goto out;
                }
        }

out:
	return i;
}

int ifp_read_seek(struct ifp_device * dev, int bytes)
{
        int i = 0;
	int available = 0;

        const int blocksize = IFP_BULK_BUFF_SIZE;
        //int bitsize   = IFP_BULK_BUFF_BITS;
        loff_t tar_offset = dev->current_offset + bytes;
        loff_t cur_offset = dev->current_offset;
        int cur_block =  cur_offset == 0 ? -1 : (int)(cur_offset-1) / blocksize;
        int tar_block =  tar_offset == 0 ? -1 : (int)(tar_offset-1) / blocksize;
        int count     = tar_block - cur_block;

	if (tar_block != -1) {
		ifp_dbg("seeking forward %d blocks from %ld to %ld (%d:%d to %d:%d)",
			count, (long)cur_offset, (long)tar_offset,
			cur_block, (int)cur_offset % blocksize,
			tar_block, (int)tar_offset % blocksize);
	}
	//ifp_dbg("seeking forward %d blocks from %ld to %ld (%d:%d to %d:%d)",
	//	count,
	//	(long)cur_offset, (long)tar_offset,
	//	cur_block, (int)cur_offset % blocksize,
	//	tar_block, (int)tar_offset % blocksize);
	IFP_BUG_ON(count < 0);

	//cases:
	//	count == 0	update current_offset and return
	//	count == 1	download next block, update current_offset
	//	count >= 2	several fake downloads, ...
	switch (count > 1 ? 2 : count) {
	case 2:
		//ifp_dbg("doing fake reads");
		i = fake_block_reads(dev, count - 1);
		ifp_err_jump(i, out, "fake block reads failed");
		//fallthrough
	case 1:
		//ifp_dbg("doing real read");
		available = (int)(dev->filesize - (tar_block*blocksize));
        if (available > blocksize) {
            available = blocksize;
        }
		i = read_next_block(dev, available);
		//ifp_err("filesize=%d, current_pos=%d; bytes=%d; available=%d",
		//	(int)dev->filesize, (int)dev->current_offset,
		//	bytes, available);
		ifp_err_jump(i, out, "error reading destination block of %d bytes",
			available);
		//fallthrough
	case 0:
		//ifp_dbg("adjusting 'current_offset' by %d", bytes);
        	dev->current_offset += bytes;
		break;
	default:
		IFP_BUG_ON(1);
	}
	//ifp_dbg("done");
	if (i) {
		ifp_dbg("returning %d", i);
	}
out:
        return i;
}

/*This code is borrowed from ifp-line and appears to work fine.  The original
 *driver behaves slightly differently:
 *
 * on open
 *	ifp_dir_open(d)
 *	ifp_file_open(f)
 *		fails because file doesn't exist yet.  Good.
 *	ifp_dir_close()
 *
 *	ifp_dir_open(d)
 *	ifp_file_open_new(d)
 *
 * on close
 *	ifp_file_close()
 *	ifp_dir_close()
 *
 *	ifp_freespace()
 *
 *	ls, I think.
 */
int ifp_write_open(struct ifp_device * dev, char * f, int n, int filesize)
{
	int i;

	IFP_BUG_ON(n > IFP_BUFFER_SIZE);
	memcpy(dev->filename, f, n);
	//memcpy(dev->dirname, d, n);

	i = ifp_copy_parent_string(dev->dirname, f, n);
	ifp_err_jump(i, out, "error copying directory name");

	i = ifp_dir_open(dev, dev->dirname, n);
	ifp_err_jump(i, out, "open_dir request failed.");

	i = ifp_freespace(dev);
	if (i < 0) {
		ifp_err_jump(i, out, "free space request failed");
	} else if (i < filesize) {
		i = -IFP_ENOSPC;
		ifp_err_jump(i, out, "not enough free space on the device");
	}

	i = ifp_file_open_new(dev, dev->filename, n, filesize);
	ifp_err_jump(i, out, "file create failed");

	IFP_BUG_ON(n > IFP_BUFFER_SIZE);
	n = n <= IFP_BUFFER_SIZE ? n : IFP_BUFFER_SIZE;
	dev->mode = IFP_MODE_WRITING;
	dev->current_offset = 0;
	dev->filesize = filesize;
	dev->readcount = 0;

out:
	return i;
}

static int quick_write_verify(struct ifp_device * dev) {
	int i = 0;
	int size;

	i = ifp_dir_open(dev, dev->dirname, IFP_BUFFER_SIZE);
	ifp_err_jump(i, out, "open dir failed");

	i = ifp_file_open(dev, dev->filename, IFP_BUFFER_SIZE);
	ifp_err_jump(i, out, "open file failed");

	size = ifp_file_size(dev);
	if (size < 0) {
		ifp_err_i(size, "file size query failed");
		goto out;
	}

	i = ifp_file_close(dev);
	ifp_err_jump(i, out, "close file failed");

	i = ifp_dir_close(dev);
	ifp_err_jump(i, out, "close dir failed");

	if (size != dev->filesize) {
		ifp_err("reported file size is %d instead of %d.. upload failed",
			size, (int)dev->filesize);
		i = IFP_EIO;
	}

out:
	return i;
}

int ifp_write_close(struct ifp_device * dev)
{
	int i = 0;
	int e = 0;
	int remainder = dev->current_offset % IFP_BULK_BUFF_SIZE;
	if (remainder != 0) {
		i = ifp_file_upload(dev, dev->iobuff, remainder);
		if (i) {
			ifp_err_i(i, "problem uploading last %d bytes.  "
			"Attempting to close file anyways.", remainder);
			e=e?e:i;
			//no jump
		}
	}
	
	ifp_wrn_on(dev->current_offset != dev->filesize,
		"received %d fewer bytes than expected",
		(int)(dev->filesize - dev->current_offset));

	//ifp_dbg("new write code finishing for %s.. all good, apparently.",
	//	dev->filename);

	i = ifp_file_flush(dev);
	if (i) {
		ifp_err_i(i, "flush failed, closing anyways.");
		e=e?e:i;
		//no jump
	}

	i = ifp_file_close(dev);
	if (i) {
		ifp_err_i(i, "file close failed, closing dir anyways.");
		e=e?e:i;
		//no jump
	}
	i = ifp_dir_close(dev);
	if (i) {
		ifp_err_i(i, "dir close failed.");
		e=e?e:i;
	}

	if (e == 0) {
		i = quick_write_verify(dev);
		if (i) {
			ifp_err_i(i, "quick-verify failed--upload may have failed.");
		}
	}
	dev->mode = IFP_MODE_NONE;

	return i?i:e;
}

int ifp_write_data(struct ifp_device * dev, void * buff, int bytes)
{
	int e = 0;
	ifp_u8 * i = buff;
	ifp_u8 * o = dev->iobuff;
	int n;
	const int bs = IFP_BULK_BUFF_SIZE;
	int block_off = dev->current_offset % bs;
	while (bytes > 0) {
		n = bs - block_off;
        if (n > bytes) {
            n = bytes;
        }
		memcpy(o + block_off, i, n);

		i += n;
		block_off += n;
		bytes -= n;
		dev->current_offset += n;

		if (block_off == bs) {
			//At the end of the file, we have to write a partial block
			e = ifp_file_upload(dev, o, block_off);
			ifp_err_jump(e, out, "upload of %d bytes failed",
				block_off);
			block_off = 0;
		}
	}
out:
	return e;
}

