/* 
 * IRiver ifp filesystem driver for Linux
 * Revision : $Id: inode.c,v 1.26 2004/09/30 00:24:27 oakhamg Exp $
 *
 * Copyright (C) Geoff Oakham, 2004; <oakhamg@users.sourceforge.net>
 */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/pagemap.h>
#include <linux/highmem.h>
#include <linux/init.h>
#include <linux/string.h>
#include <linux/statfs.h>
#include <linux/smp_lock.h>
#include <linux/backing-dev.h>
#include <linux/usb.h>
#include <linux/nls.h>
#include <linux/writeback.h>
#include <linux/mpage.h>
#include <linux/parser.h>

#include <asm/uaccess.h>

#include "iriverfs.h"
#include "ifp.h"
#include "ifp_dentry.h"
#include "ifp_os.h"
#include "ifp_prim.h"

/* some random numbers */
#define IRIVERFS_MAGIC	0x63880F53


#if 1
//I don't know why, but blocksize appears to be locked at 512.  I'd very
//much like to set it to 16384, but until I can get userland 'stat' to see
//this, I'm going to stick with 512.
#define IFP_BLOCK_SIZE_FAKE     512
#define IFP_BLOCK_BITS_FAKE     9
#else
#define IFP_BLOCK_SIZE_FAKE     16384
#define IFP_BLOCK_BITS_FAKE     14
#endif

static struct super_operations iriver_ops;
//static struct address_space_operations iriver_aops;
static struct file_operations iriver_file_operations;
static struct file_operations iriver_dir_operations;
static struct inode_operations iriver_file_inode_operations;
static struct inode_operations iriver_dir_inode_operations;

static int iriver_readpage(struct file *file, struct page * page);

static int put_ifp_device(struct ifp_device * dev, int disconnected);
//static int iriver_dbg_writepages(struct address_space *mapping, struct writeback_control *wbc) ;

static int readpage_error(struct file *file, struct page * page);

//romfs version
static struct address_space_operations aops_restricted = {
	.readpage	= readpage_error,
};

static struct address_space_operations aops_cache = {
	.readpage	= iriver_readpage,
};

//used in ramfs
static struct address_space_operations aops_ram = {
	.readpage	= simple_readpage,
	.prepare_write	= simple_prepare_write,
	.commit_write	= simple_commit_write
};


//The various states an inode can be in.  These reflect 
#define IFP_I_NORMAL 	0
#define IFP_I_NEW	1
#define IFP_I_DIRTY	2 /* potentially dirty.. check individual pages */
struct iriver_inode_info {
	int loaded; //[dir only] true if the directory has been read into dcache
	int state; //[file only] true when a file doesn't yet exist on disk

	struct inode vfs_inode;
};
static inline struct iriver_inode_info *IRIVER_I(struct inode *inode) {
	return list_entry(inode, struct iriver_inode_info, vfs_inode);
}

#define PAGES_PER_BUFFER (IFP_BULK_BUFF_SIZE/PAGE_CACHE_SIZE)
#if IFP_BULK_BUFF_SIZE < PAGE_CACHE_SIZE || IFP_BULK_BUFF_SIZE % PAGE_CACHE_SIZE
	//This is an assumption that makes the driver easier to write.  I know
	//it won't necessarily hold for all platforms, so the best practice
	//seems to be nice: abort at compile time when the assumption fails.
	#error "This driver doesn't support this system's page size."
#endif

struct iriver_readstate {
	struct dentry * dentry;
	struct inode * inode;
	long currentpage;
};

struct iriver_mount_options {
	uid_t uid;
	gid_t gid;
	unsigned short fmask;
	unsigned short dmask;
};

struct iriver_sb_info {
	struct ifp_device * dev;
	struct semaphore lock;

	struct iriver_mount_options options;

	//"Lazy cleanup" values that allow the device to be left in unusual
	//states, so long as the cu^H^Hfunction leaves a 'cleanup()'
	//function here.
	//
	//Cleanup happens (if necessary) when the lock is aquired.
	void (*cleanup)(struct super_block *, void * state);
	void * cleanup_state;

	struct iriver_readstate readstate;
	
	//True normally, false during mount, unmount or disconnect.
	int active;
};
static inline struct iriver_sb_info *IRIVER_S(struct super_block *sb) {
	return sb ? (struct iriver_sb_info *)sb->s_fs_info : NULL;
}

//Used in RAMFS
static struct backing_dev_info iriver_backing_dev_info = {
	.ra_pages	= 0,	/* No readahead */
	.capabilities	= BDI_CAP_NO_ACCT_DIRTY | BDI_CAP_NO_WRITEBACK |
			  BDI_CAP_MAP_DIRECT | BDI_CAP_MAP_COPY |
			  BDI_CAP_READ_MAP | BDI_CAP_WRITE_MAP | BDI_CAP_EXEC_MAP,
};

static inline void cleanup(struct super_block *sb) {
	if (IRIVER_S(sb)->cleanup) {
		IRIVER_S(sb)->cleanup(sb, IRIVER_S(sb)->cleanup_state);
		IRIVER_S(sb)->cleanup = IRIVER_S(sb)->cleanup_state = NULL;
	}
}

static inline int _ifp_lock(struct super_block *sb) {
	down(&IRIVER_S(sb)->lock);
	if (IRIVER_S(sb)->dev == NULL) {
		printk("[ifp_lock] device disconnected, apparently.\n");
		up(&IRIVER_S(sb)->lock);
		return -EPIPE;
	}
	BUG_ON(IRIVER_S(sb)->active == 0);
	//BUG_ON(IRIVER_S(sb)->dev == NULL);
	cleanup(sb);
	return 0;
}

#define ifp_lock(a) if(_ifp_lock(a)) { \
	printk("[%s] returning\n", __FUNCTION__); return -EPIPE;}

static inline void ifp_unlock(struct super_block *sb) {
	up(&IRIVER_S(sb)->lock);
}

static inline int _ifp_debug_msg(char const * fn,
		struct super_block * sb,
		struct dentry * dentry,
		char * message)
{
	int i = 0;
	struct dentry * root = sb->s_root;

	char * z = get_iriver_buffer();
	if (z == NULL) {
		printk("[_ifp_debug_msg] z is NULL\n");
		put_iriver_buffer(z); z = NULL;
		return -ENOMEM;
	}
	if (root == NULL) {
		printk("[_ifp_debug_msg] root is NULL\n");
		put_iriver_buffer(z); z = NULL;
		return -1;
	}
	i = make_path_string(dentry, root, '/', z, IFP_BUFFER_SIZE);
	if (i < 0) {
		printk("[_ifp_debug_msg] i is %d\n", i);
	}
	printk("[%20.20s] ifp:/%s %s\n", fn, z, message);
	put_iriver_buffer(z); z = NULL;

	return i;
}

#ifdef IFP_DEBUG_SYSCALL
#define ifp_debug_msg(s, d, msg) _ifp_debug_msg(__FUNCTION__, s, d, msg)
#else
#define ifp_debug_msg(s, d, msg) 
#endif


//-------------------------------------------------------------------------
//inode and dcache manipulation

static struct inode *iriver_get_inode(struct super_block *sb, int mode, dev_t dev)
{
	struct inode * inode = new_inode(sb);

	if (inode) {
		inode->i_mode = mode;
		inode->i_uid = IRIVER_S(sb)->options.uid;
		inode->i_gid = IRIVER_S(sb)->options.gid;
		inode->i_blkbits = IFP_BLOCK_BITS_FAKE;
		inode->i_blksize = IFP_BLOCK_SIZE_FAKE;
		inode->i_blocks = 0;
		inode->i_mapping->a_ops = &aops_cache;
		inode->i_mapping->backing_dev_info = &iriver_backing_dev_info;
		inode->i_atime = inode->i_mtime = inode->i_ctime = CURRENT_TIME;
		inode->i_flags |= S_NOATIME; //S_APPEND
		switch (mode & S_IFMT) {
		default:
			init_special_inode(inode, mode, dev);
			break;
		case S_IFREG:
			inode->i_op = &iriver_file_inode_operations;
			inode->i_fop = &iriver_file_operations;
			break;
		case S_IFDIR:
			inode->i_size = IFP_BLOCK_SIZE_FAKE;
			inode->i_blocks = 1;
			inode->i_op = &iriver_dir_inode_operations;
			inode->i_fop = &iriver_dir_operations;

			/* directory inodes start off with i_nlink == 2 (for "." entry) */
			inode->i_nlink++;
			break;
		case S_IFLNK:
			inode->i_op = &page_symlink_inode_operations;
			break;
		}
	}
	return inode;
}

static int
iriver_mknod(struct inode *dir, struct dentry *dentry, int mode, dev_t dev)
{
	struct inode * inode = iriver_get_inode(dir->i_sb, mode, dev);
	int error = -ENOSPC;

	if (inode) {
		//Problem: calling the two lines labled "suspect" causes crashes
		//on the first flushd after writing a file.  However, leaving
		//them out causes newly created directories to not appear.
        
		//insert_inode_hash(inode); //suspect
		if (dir->i_mode & S_ISGID) {
			inode->i_gid = dir->i_gid;
			if (S_ISDIR(mode))
				inode->i_mode |= S_ISGID;
		}
		d_instantiate(dentry, inode);
		//d_rehash(dentry); //suspect
		dget(dentry);	/* Extra count - pin the dentry in core */
		error = 0;
	}
	return error;
}

static int iriver_mkdir(struct inode * dir, struct dentry * dentry, int mode)
{
	int retval;

	ifp_debug_msg(dir->i_sb, dentry, "");
	
	ifp_lock(dir->i_sb);
	retval = ifp_mkdir_d(IRIVER_S(dir->i_sb)->dev, dentry);
	if (retval) {
		printk("[iriver_mkdir] error %d.\n", retval);
		if (retval < 0) {
			ifp_unlock(dir->i_sb);
			return retval;
		} else {
			ifp_unlock(dir->i_sb);
			return -EIO;
		}
	}

	//override 'mode'
	mode = (S_IRWXUGO & ~IRIVER_S(dir->i_sb)->options.dmask);
	retval = iriver_mknod(dir, dentry, mode | S_IFDIR, 0);
	if (retval) {
		printk("[iriver_mkdir] mknod return an error, %d.\n", retval);
	}
	//printk("[iriver_mkdir] nlink was %d, ", dir->i_nlink);
	if (!retval) {
		//new directories are empty
		IRIVER_I(dentry->d_inode)->loaded = 1;
		dir->i_nlink++;
	}
	//printk("now %d\n", dir->i_nlink);

	ifp_unlock(dir->i_sb);
	return retval;
}

static int create_new_file(struct inode *dir, struct dentry *dentry, int mode,
	struct nameidata *nd)
{
	int e;
	ifp_debug_msg(dir->i_sb, dentry, "");
	ifp_lock(dir->i_sb);

	//override 'mode'
	mode = (S_IRWXUGO & ~IRIVER_S(dir->i_sb)->options.dmask);
	e = iriver_mknod(dir, dentry, mode | S_IFREG, 0);
	if (!e) {
		IRIVER_I(dentry->d_inode)->state = IFP_I_NEW;
		
		dentry->d_inode->i_mapping->a_ops = &aops_ram;

		//inode->i_mapping->aops =  &aops_cache
		//dentry->d_inode->i_mapping->backing_dev_info = &iriver_backing_dev_info;
		//dentry->d_inode->i_mapping->backing_dev_info = NULL;
		//dentry->d_inode->i_mapping->a_ops = &aops_cache;
	}
	ifp_unlock(dir->i_sb);
	return e;
}

struct fill_dir_simple_context {
	struct dentry * parent;

	//option 1
	//Can be NULL
	struct dentry * requested;
	struct inode *  found;
};

static struct dentry * create_normal_dentry(
	struct fill_dir_simple_context * context,
	char * filename,
	int n,
	int mode,
	int restricted,
	struct inode ** pinode
	)
{
	struct qstr s;
	struct dentry * d = NULL;
	//struct inode * inode;

	s.name = filename;
	s.len = n;
	s.hash = full_name_hash(s.name, s.len);

	*pinode = iriver_get_inode(context->parent->d_sb, mode, 0);
	if (*pinode == NULL) {
		printk("[create_normal_dentry] iriver_get_inode failed.\n");
		return NULL;
	}

	if (context->requested
		&& s.len == context->requested->d_name.len
		&& 0 == memcmp(context->requested->d_name.name, s.name, s.len))
	{
		//printk("[create_normal_dentry] matched '%s'!\n", s.name);
		d = context->requested;
		context->found = *pinode;
	} else {
		d = d_alloc(context->parent, &s);
		if (d == NULL) {
			//FIXME: free 'inode', if we allocated it.  (Is that possible?)
			printk("[create_normal_dentry] d_alloc failed.\n");
			return NULL;
		}
		insert_inode_hash(*pinode);
		d_add(d, *pinode);  // == d_instantiate + d_rehash
	}

	if (restricted) {
		(*pinode)->i_mapping->a_ops = &aops_restricted;
	}

	return d;
}

static int iriver_symlink(struct inode * dir, struct dentry *dentry,
	const char * symname)
{
	ifp_debug_msg(dir->i_sb, dentry, "");
	printk("iRiverFS: links (symbolic or otherwise) are not supported.\n");
	return -EPERM;
}

static int init_inode_regular(struct ifp_device * dev, struct super_block * sb,
	struct inode * inode, struct dentry * d, int filesize)
{
	inode->i_size = filesize;

	inode->i_blksize = IFP_BLOCK_SIZE_FAKE;
	inode->i_blkbits = IFP_BLOCK_BITS_FAKE;
	//inode->i_blocks = 1 + filesize/(inode->i_blksize);
	inode->i_blocks = filesize >> inode->i_blkbits;

	return 0;
}

int iriver_revalidate(struct dentry * d, struct nameidata * nameidata);

static struct dentry_operations dop_validate = {
	.d_revalidate = iriver_revalidate,
};

static int fill_dir_simple(void * state, int type, char * filename, int n,
	int filesize)
{
	int i = 0;
	int mode;
	struct dentry * d = NULL;
	struct inode * inode = NULL;

	struct fill_dir_simple_context * context = state;
	struct dentry * parent = context->parent;
	struct super_block * sb = context->parent->d_sb;

	struct ifp_device * dev = IRIVER_S(sb)->dev;
	int restricted = 0;

	if (type == IFP_FILE) {
		int mask = IRIVER_S(sb)->options.fmask;
		//later models to not have restricted file downloads.
		if (dev->model <= 0x1005 && n > 4) {
			//This limitation of the hardware; all attempts
			//to read 'restricted files' return junk.
			//If you know of a work-around, please let
			//me know!
			char * sz = filename + n - 4;
			if (strnicmp(sz, ".mp3", 4) == 0
				|| strnicmp(sz, ".wma", 4) == 0
				|| strnicmp(sz, ".asf", 4) == 0)
			{
				restricted = 1;
				mask |= S_IRUGO;
				//This is probably more for ethstetics
				//mask |= S_IWUGO;
			}
		}
		//until the driver supports file modification, this seems like
		//the sensible thing.  Unfortunately some shells prompt the
		//user before deleting non-writable files.. so perhaps this
		//isn't the best of ideas.
		//mask |= S_IWUGO;

		mode = (S_IRWXUGO & ~mask) | S_IFREG;
	} else if (type == IFP_DIR) {
		mode = (S_IRWXUGO & ~IRIVER_S(sb)->options.dmask) | S_IFDIR;
	} else {
		printk("[fill_dir_simple] I don't know what to do with type = %d, stopping\n", type);
		return -EIO;
	}

	d = create_normal_dentry(context, filename, n, mode, restricted,
		&inode);
	//inode = d->d_inode;

	if (type == IFP_FILE) {
		//printk("file [ino=%ld]: len=%d '%s'\n", inode->i_ino, len,
		//	utf8_buf);
		i = init_inode_regular(IRIVER_S(sb)->dev, sb, inode, d,
			filesize);
		if (i) {
			printk("[callbk] read_inode_regular(ino=%ld, '%s') returned error %d, bailing\n", inode->i_ino, "" /*s.name*/, i);
			return i;
		}
		IRIVER_I(inode)->state = IFP_I_NORMAL;
	} else if (type == IFP_DIR) {
		//printk("dir  [ino=%ld]: len=%d '%s'\n", inode->i_ino, len,
		//	utf8_buf);
		parent->d_inode->i_nlink++;
		d->d_op = &dop_validate;
	}

	return i;
}

static int preload_directory(struct ifp_device * dev, struct dentry * d,
	struct dentry * d_request, struct inode ** pinode_return)
{
	int i = 0;
	struct fill_dir_simple_context context;
	context.parent = d;
	context.requested = d_request;
	context.found = NULL;

	if (IRIVER_I(d->d_inode)->loaded != 0) {
		printk("[preload_directory] attempting to preload something that's allready preloaded.. loaded=%d\n",
			IRIVER_I(d->d_inode)->loaded);
	}
	ifp_debug_msg(d->d_sb, d, "loading");
	i = ifp_list_dirs_d(dev, d, fill_dir_simple, &context);
	if (i) {
		printk("[preload_directory] err %d\n",i);
		return i;
	}
	if (d_request && pinode_return) {
		*pinode_return = context.found;
	}

	IRIVER_I(d->d_inode)->loaded = 1;
	d->d_op = &dop_validate;

	return i;
}

static int load_directory(struct ifp_device * dev, struct dentry * d,
	struct dentry * d_request, struct inode ** pinode_return)
{
	int i = 0;
	if (IRIVER_I(d->d_inode)->loaded == 0) {
		i = preload_directory(IRIVER_S(d->d_sb)->dev, d, d_request,
			pinode_return);
		if (i) {
			printk("[iriver_readdir] preload error, i=%d.\n",i);
			goto out;
		}
	} else {
		//printk("[load_directory] doing nothing because loaded=%d\n",
		//	IRIVER_I(d->d_inode)->loaded);
	}

out:
	return i;
}

int iriver_revalidate(struct dentry * d, struct nameidata * nameidata)
{
//	int i = 0;
	if (d && d->d_inode) {
		//struct super_block * sb = d->d_sb;
#if 1
		ifp_debug_msg(d->d_sb, d, "doing nothing");
#else
		ifp_debug_msg(sb, d, "");

		i = load_directory(IRIVER_S(sb)->dev, d, NULL, NULL);
		if (i) {
			printk("[iriver_revalidate] error loading directory, err = %d.\n",i);
			return i>0 ? -1 : i;
		}
#endif
		//stop ourselfs from being called--this inode should now be
		//fully loaded.
		d->d_op = NULL;
	} else {
		printk("[iriver_revalidate] here but something's null.\n");
	}

	return 1;
}

static int iriver_readdir(struct file *filp, void *dirent, filldir_t filldir)
{
	int i = 0;
	struct dentry *dentry = filp->f_dentry;
	struct super_block * sb = dentry->d_sb;
	struct inode *inode;

	ifp_lock(sb);

	inode = dentry->d_inode;

#if 0
	if (IRIVER_I(dentry->d_inode)->loaded != 1) {
		ifp_debug_msg(sb, dentry, "");
	}
#endif
	ifp_debug_msg(sb, dentry, "");
	i = load_directory(IRIVER_S(sb)->dev, dentry, NULL, NULL);
	if (i) {
		printk("[iriver_readdir] load directory failed, i = %d.\n",i);
		if (i > 0) {
			i = -1;
		}
		goto out;
	}

	i = dcache_readdir(filp, dirent, filldir);
out:
	ifp_unlock(sb);
	return i;
}

static struct dentry * iriver_lookup(struct inode * i_parent,
	struct dentry * dentry, struct nameidata * namei)
{
	struct super_block * sb = i_parent->i_sb;
	struct dentry * dreturn;
	struct ifp_device * dev = IRIVER_S(sb)->dev;
	int i = 0;

	if (_ifp_lock(sb)) {
		printk("[%s] returning 'cause I couldn't get a lock\n", __FUNCTION__);
		d_add(dentry, NULL);
		return ERR_PTR(-EPIPE); 
	}

	if (IRIVER_I(i_parent)->loaded != 1) {
		struct inode * found_inode = NULL;
		ifp_debug_msg(i_parent->i_sb, dentry, "");
		i = load_directory(dev, dentry->d_parent, dentry,
			&found_inode);
		if (i) {
			printk("[iriver_lookup] load directory failed,i=%d.\n",
				i);
			goto out_novalue;
		}
		if (found_inode) {
			struct dentry * dd = NULL;
			dreturn = d_splice_alias(found_inode, dentry);
			if (dreturn) {
				dd = dreturn;
			} else {
				dd = dentry;
			}

			atomic_inc(&dd->d_count);
			goto out;
		} else {
			d_add(dentry, NULL);
		}
	} else {
		ifp_debug_msg(i_parent->i_sb, dentry, "...none");
		if (dentry->d_name.len > NAME_MAX) {
			return ERR_PTR(-ENAMETOOLONG);
		}
		d_add(dentry, NULL);
		i = 0;
	}

out_novalue:
	dreturn = ERR_PTR(i);
out:
	ifp_unlock(sb);

	return dreturn;
}

//-------------------------------------------------------------------------
//misc operations (df, rmdir, unlink, rename.. etc)
static int iriver_statfs(struct super_block * sb, struct kstatfs * buf)
{
	int r;
	unsigned int capacity, free;

#ifdef IFP_DEBUG_SYSCALL
	printk("[iriver_statfs] here\n");
#endif
	ifp_lock(sb);

	r = simple_statfs(sb, buf);
	BUG_ON(r);

	capacity = ifp_capacity(IRIVER_S(sb)->dev);
	if (capacity < 0) { printk("iriver_statfs: error requesting capactity, err %d\n",
			capacity);
		ifp_unlock(sb);
		return -EIO;
	}
	free = ifp_freespace(IRIVER_S(sb)->dev);
	if (free < 0) {
		printk("iriver_statfs: error requesting ammount free, err %d\n",
			free);
		ifp_unlock(sb);
		return -EIO;
	}
	ifp_unlock(sb);

	buf->f_bsize = IFP_BLOCK_SIZE_FAKE;
	buf->f_namelen = IFP_BULK_MAXPATHLEN;

	buf->f_blocks = capacity/buf->f_bsize; /* ok ? */
	buf->f_bfree = free/buf->f_bsize; /* not right */

	buf->f_bavail = buf->f_bfree;

	return 0;
}

int iriver_dir_getattr(struct vfsmount *mnt, struct dentry *d,
	struct kstat *stat)
{
	int i;
	ifp_lock(d->d_sb);

	ifp_debug_msg(d->d_sb, d, "");
	if (IRIVER_I(d->d_inode)->loaded == 0) {
		i = preload_directory(IRIVER_S(d->d_sb)->dev, d, NULL, NULL);
		if (i) {
			printk("[iriver_getattr] preload error, i=%d.\n",i);
			goto out;
		}
	}

	i = simple_getattr(mnt, d, stat);

out:
	ifp_unlock(d->d_sb);
	return i;
}

int iriver_file_getattr(struct vfsmount *mnt, struct dentry *d,
	struct kstat *stat)
{
	int i;
	ifp_lock(d->d_sb);

	i = simple_getattr(mnt, d, stat);

	ifp_debug_msg(d->d_sb, d, "");
	stat->blksize = IFP_BLOCK_SIZE_FAKE;
	stat->blocks = 1 + (stat->size >> IFP_BLOCK_BITS_FAKE);
	//ifp_dbg("setting as %d blocks of %d, for %d bytes total",
	//	(int)stat->blocks, (int)stat->blksize, (int)stat->size);

	ifp_unlock(d->d_sb);
	return i;
}

int iriver_setattr(struct dentry * d, struct iattr * attr)
{
	int i = 0;
	struct inode * inode = d->d_inode;
	struct super_block * sb = inode->i_sb;

	ifp_debug_msg(d->d_sb, d, "");
	lock_kernel();

	//I can't handle truncation or expansion
	if (attr->ia_valid & ATTR_SIZE) {
		i = -EPERM;
		goto out;
	}

	i = inode_change_ok(inode, attr);
	if (i)
		goto out;

	if (((attr->ia_valid & ATTR_UID) &&
		 (attr->ia_uid != IRIVER_S(sb)->options.uid)) ||
		((attr->ia_valid & ATTR_GID) &&
		 (attr->ia_gid != IRIVER_S(sb)->options.gid)) ||
		 (attr->ia_valid & ATTR_MODE) ||
		 (attr->ia_valid & ATTR_ATIME_SET) ||
		 (attr->ia_valid & ATTR_MTIME_SET)
		 )
	{
		 //i = -EPERM;
		 i = 0;
		 goto out;
	}

	//does this step do anything?
	i = inode_setattr(inode, attr);
	if (i) 
		goto out;
out:
	unlock_kernel();
	return i;
}

static struct file_operations iriver_dir_operations = {
	.open           = dcache_dir_open,
	.release        = dcache_dir_close,
	.llseek         = dcache_dir_lseek,
	.read		= generic_read_dir,

	.readdir        = iriver_readdir,
};

static struct inode_operations iriver_file_inode_operations = {
	//should probably write my own.
	.getattr	= iriver_file_getattr,
	.setattr	= iriver_setattr,
};


static int iriver_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry)
{
	printk("iRiverFS: hard links are not supported.\n");
	return -EPERM;
}

static int iriver_unlink(struct inode *dir, struct dentry *dentry)
{
	int i;
	ifp_debug_msg(dir->i_sb, dentry, "");
	ifp_lock(dir->i_sb);
	i = ifp_delete_d(IRIVER_S(dir->i_sb)->dev, dentry);
	if (i) {
		printk("[iriver_unlink] error %d.\n", i);
		if (i > 0) {
			i = -EIO;
		}
		goto out;
	}

	i = simple_unlink(dir, dentry);
out:
	ifp_unlock(dir->i_sb);
	return i;
}

static int iriver_rmdir(struct inode *dir, struct dentry *dentry)
{
	int i;
	ifp_debug_msg(dir->i_sb, dentry, "");
	ifp_lock(dir->i_sb);
	i = ifp_rmdir_d(IRIVER_S(dir->i_sb)->dev, dentry);
	ifp_unlock(dir->i_sb);
	if (i) {
		printk("[iriver_rmdir] error %d.\n", i);
		if (i > 0) {
			i = -EIO;
		}
		goto out;
	}
	i = simple_rmdir(dir, dentry);
out:
	//ifp_unlock(dir->i_sb);
	return i;
}

static int iriver_rename(struct inode *old_dir, struct dentry *old_dentry,
	struct inode *new_dir, struct dentry *new_dentry)
{
	struct ifp_device * dev = IRIVER_S(old_dir->i_sb)->dev;
	int i = 0;
	ifp_debug_msg(old_dir->i_sb, old_dentry, "");
	//printk("iRiverFS warning: renaming files is experimental.\n");

#ifdef IFP_RENAME
	if (S_ISDIR(old_dentry->d_inode->i_mode)) {
		printk("iRiverFS warning: moving directories is relatively new.\n");
	}
#if 0
	if (old_dir != new_dir) {
		printk("iRiverFS warning: moving files to different directories is even more experimental.\n");
	}
#endif
	ifp_lock(old_dir->i_sb);
	i = ifp_rename_d(dev, old_dentry, new_dentry);
	ifp_unlock(old_dir->i_sb);

	if (i) {
		printk("[iriver_rename] rename failed, err %d\n", i);
		i = -EIO;
		goto out;
	}
out:
#else
	printk("iRiverFS error: renaming support is disabled.\n");
	i = -EPERM;
#endif
	return i;
}

static int
iriver_mknod_EIO(struct inode *dir, struct dentry *dentry, int mode, dev_t dev) {
	printk("iRiverFS: special devices are not supported.\n");
	return -EPERM;
}

//-------------------------------------------------------------------------
//uploading files
static int iriver_open(struct inode * inode, struct file * f)
{
	int i = 0;
	ifp_debug_msg(inode->i_sb, f->f_dentry, "");
	ifp_lock(inode->i_sb);

	//If the state's allready NEW or DIRTY, leave it.
	if (IRIVER_I(inode)->state == IFP_I_NORMAL) {
		if ((f->f_mode & FMODE_WRITE) == O_RDONLY) {
			//printk("[iriver_open] Opening read-only...\n");
			if (inode->i_mapping->a_ops == &aops_restricted) {
				printk("iRiverFS: warning, access to some files is restricted by firmware.\n");
				i = -EPERM;
				goto out;
			}
		} else {
			printk("iRiverFS warning: editing files is currently not supported.\n");
			i = -EPERM;
			goto out;
#if 0
			IRIVER_I(inode)->state = IFP_I_DIRTY;
			inode->i_mapping->a_ops = &aops_ram;
			inode->i_mapping->backing_dev_info = &iriver_backing_dev_info;
			printk("[iriver_open] Opening for writing...\n");
#endif
		}
	} else if (IRIVER_I(inode)->state == IFP_I_NEW) {
		//printk("[iriver_open] Opening a new file...\n");
	} else {
		printk("[iriver_open] unknown state %d.\n", IRIVER_I(inode)->state);
		i = -EIO;
	}

out:
	ifp_unlock(inode->i_sb);
	return i;
}

static int iriver_write_new_file(struct inode * inode, struct file * f)
{
	struct ifp_device * dev = IRIVER_S(inode->i_sb)->dev;
	long bytes = i_size_read(inode);
	struct page * page = NULL;
	void * pgdata = NULL;
	int pagenum;
	int i = 0, e = 0;

	i = ifp_write_open_d(dev, f->f_dentry, inode->i_size);
	if (i) {
		printk("[iriver_write_new_file] start failed, err %d\n",i);
		return i;
	}

	pagenum = 0;
	while (bytes > 0)
	{
		int n = (bytes < PAGE_CACHE_SIZE) ? bytes : PAGE_CACHE_SIZE;

		page = find_lock_page(inode->i_mapping, pagenum++);
		pgdata = kmap(page);
		BUG_ON(page == NULL);
		BUG_ON(pgdata == NULL);

		i = ifp_write_data(dev, pgdata, n);

		flush_dcache_page(page);
		kunmap(page);
		unlock_page(page);
		page_cache_release(page);
		page = NULL;

		if (i) {
			ifp_err_i(i, "error writing page of data");
			e = e?e:i;
			break;
		}

		bytes -= n;
	}
	BUG_ON(bytes != 0);

	i = ifp_write_close(dev);
	if (i) {
		printk("[iriver_write_new_file] end failed, err %d\n",i);
	}

	return e?e:i;
}


static int write_file(struct inode * inode, struct file * f) {
	int i;
	
	//FIXME: if this fails, either try again or nuke the inode & dentry.
	i = iriver_write_new_file(inode, f);
	if (i) {
		printk("[write_file] failed..\n");
		return i;
	}

	//printk("[write_file] File upload sucessful.\n");
	IRIVER_I(inode)->state = IFP_I_NORMAL; 
	inode->i_mapping->a_ops =  &aops_cache;

	return i;
}

static int iriver_release(struct inode * inode, struct file * f) {
	int i = 0;
	ifp_lock(inode->i_sb);
	if (IRIVER_I(inode)->state == IFP_I_NEW) {
		//printk("[iriver_release] Creating file.  (Blocking while we upload).\n");

		ifp_debug_msg(inode->i_sb, f->f_dentry, "writing file (blocking).");
		i = write_file(inode, f);
		if (i) {
			printk("[iriver_release] file write failed.\n");
			goto out;
		}
	} else if (IRIVER_I(inode)->state == IFP_I_DIRTY) {
		ifp_debug_msg(inode->i_sb, f->f_dentry, "");
		printk("[iriver_release] Existing file that might need updating.\n");
	}
out:
	ifp_unlock(inode->i_sb);
	return i;
}

//IMPROVEMENT IDEA: make two versions of this, one for new
//files and one for existing files.  This eliminates the need 
//for 'state' flags.
static struct file_operations iriver_file_operations = {
	.read		= generic_file_read,
	.write		= generic_file_write,
	.mmap		= generic_file_mmap,
	.fsync		= simple_sync_file,
	.sendfile	= generic_file_sendfile,
	.llseek		= generic_file_llseek,

	.open		= iriver_open,
	.release	= iriver_release,
};


//-------------------------------------------------------------------------
//downloading files

//cheating_page
static void ifp_download_cleanup(struct super_block * sb,
	void * vp)
{
	int i;
	struct iriver_readstate * p = vp;

	if (IRIVER_S(sb) == NULL || IRIVER_S(sb)->dev == NULL) {
		printk("[ifp_download_cleanup] one of them is NULL.\n");
	}

	//ifp_debug_msg needs access to dentrys that might allready be
	//free().  But, we don't actually need those dentries to cleanup
	//the ifp's device state.. so this is the comprimise.
	if (IRIVER_S(sb)->active) {
		//ifp_debug_msg(sb, p->dentry, "");
	} else {
		printk("[ifp_download_cleanup] here..\n");
	}
	i = ifp_read_close(IRIVER_S(sb)->dev);
	if (i) {
		printk("[ifp_download_cleanup] file leave control message failed (err %d).\n", i);
	}

	p->currentpage = 0;
	p->inode = NULL;
	p->dentry = NULL;
}

static int
readpage_error(struct file *file, struct page * page)
{
	ifp_debug_msg(page->mapping->host->i_sb, file->f_dentry, "");
	printk("iRiverFS: warning, access to some files is restricted by firmware.\n");
	SetPageError(page);
	unlock_page(page);
	return -EPERM;
}

static int
iriver_readpage(struct file *file, struct page * page)
{
	struct inode * inode = page->mapping->host;
	u32 maxblock, bytes_filled;
	void * pgdata;
	int i = 0;
	struct ifp_device * dev = IRIVER_S(inode->i_sb)->dev;
	struct super_block *sb = inode->i_sb;

	maxblock = (inode->i_size + PAGE_CACHE_SIZE - 1) >> PAGE_CACHE_SHIFT;
	bytes_filled = 0;
	//ifp_debug_msg(sb, file->f_dentry, "reading page");
	//ifp_dbg("reading page %d", (int)page->index);

	if (page->index < maxblock) {
		down(&IRIVER_S(sb)->lock);
		if (IRIVER_S(sb)->dev == NULL) {
			printk("[iriver_readpage] device disconnected");
			i = -EPIPE;
			goto err_2;
		}
		//printk("[iriver_readpage] starting for page %d (size=%d, maxp=%d)\n",
		//	(int)page->index, (int)inode->i_size, (int)maxblock);

		if (IRIVER_S(sb)->cleanup == ifp_download_cleanup
			&& IRIVER_S(sb)->cleanup_state == 
				(void*)&(IRIVER_S(sb)->readstate)

			&& IRIVER_S(sb)->readstate.currentpage == page->index)
		{
			IRIVER_S(sb)->cleanup = NULL;
			IRIVER_S(sb)->cleanup_state = NULL;
		} else {
			if (IRIVER_S(sb)->cleanup) {
				IRIVER_S(sb)->cleanup(sb, IRIVER_S(sb)->cleanup_state);
				IRIVER_S(sb)->cleanup
					= IRIVER_S(sb)->cleanup_state
					= NULL;
				IRIVER_S(sb)->readstate.currentpage = 0;
			}

			i = ifp_read_open_d(dev, file->f_dentry);
			if (i) {
				printk("[iriver_readpage] ifp_read_open_d %d.\n", i);
				i = -EIO;
				goto err_2;

			}

			//printk("[iriver_readpage] seeking to %d\n",
			//	(int)page->index);
			i = ifp_read_seek(dev, PAGE_CACHE_SIZE*page->index);
			if (i) {
				printk("[iriver_readpage] ifp_read_seek returned %d.\n", i);
				IRIVER_S(sb)->cleanup = ifp_download_cleanup;
				IRIVER_S(sb)->cleanup_state = &IRIVER_S(sb)->readstate;
				i = -EIO;
				goto err_1;
			}
		}

		ClearPageUptodate(page);
		pgdata = kmap(page);
		i = ifp_read_data(dev, pgdata, PAGE_CACHE_SIZE);
		if (i < 0) {
			printk("[iriver_readpage] read_fileblock failed with err %d.\n", i);
			IRIVER_S(sb)->cleanup = ifp_download_cleanup;
			IRIVER_S(sb)->cleanup_state = &IRIVER_S(sb)->readstate;
			kunmap(page);
			goto err_1;
		} else if (i < PAGE_CACHE_SIZE) {
			//printk("[iriver_readpage] got %d instead of %d.\n", i, (int)PAGE_CACHE_SIZE);
			memset(pgdata + i, 0, PAGE_CACHE_SIZE-i);
			i = 0;
		} else {
			i = 0;
		}

		flush_dcache_page(page);
		SetPageUptodate(page);

		IRIVER_S(sb)->readstate.dentry = file->f_dentry;
		IRIVER_S(sb)->readstate.inode = inode;
		IRIVER_S(sb)->readstate.currentpage = page->index + 1;

		IRIVER_S(sb)->cleanup = ifp_download_cleanup;
		IRIVER_S(sb)->cleanup_state = &IRIVER_S(sb)->readstate;
		up(&IRIVER_S(sb)->lock);
	} else {
		//printk("[iriver_readpage] empty page over here.. page %d\n",
		//	(int)page->index);
		pgdata = kmap(page);
		memset(pgdata, 0, PAGE_CACHE_SIZE);
		SetPageUptodate(page);
	}

	unlock_page(page);
	kunmap(page);

	return 0;
	
err_1:
	ifp_download_cleanup(sb, IRIVER_S(sb)->cleanup_state);
	IRIVER_S(sb)->cleanup = IRIVER_S(sb)->cleanup_state
		= NULL;
err_2:
	up(&IRIVER_S(sb)->lock);
	SetPageError(page);
	if (i) {
		ifp_err_i(i, "returning error for page %d", (int)page->index);
	} else {
		ifp_err("*not* returning error for page %d.. wierd", (int)page->index);
	}
	unlock_page(page);

	return i;
}


static struct inode_operations iriver_dir_inode_operations = {
	.create		= create_new_file,
	.lookup		= iriver_lookup,
	.link		= iriver_link, //simple_link,
	.unlink		= iriver_unlink, //simple_unlink,
	.symlink	= iriver_symlink,
	.mkdir		= iriver_mkdir,
	.rmdir		= iriver_rmdir, //simple_rmdir,
	.mknod		= iriver_mknod_EIO,
	.rename		= iriver_rename, //simple_rename,
	.getattr	= iriver_dir_getattr,
	.setattr	= iriver_setattr,
};

//-------------------------------------------------------------------------
//setup, initialization and cleanup
static int driver_probe (struct usb_interface *intf,
	const struct usb_device_id *id)
{
        return -ENODEV;
}

static void driver_disconnect(struct usb_interface *intf)
{
	struct super_block * sb = (struct super_block *)usb_get_intfdata(intf);
	if (sb == NULL) {
		printk("[driver_disconnect] sb is NULL.  returning.\n");
        return;
	}
	if (sb->s_fs_info == NULL) {
		printk("[driver_disconnect] sb->s_fs_info is NULL.  returning.\n");
		return;
	}
	if (IRIVER_S(sb)->active) {
		int i = 0;
		struct usb_device * device_handle = IRIVER_S(sb)->dev->private;

		printk("[driver_disconnect] unplugged!\n");

		//The hope is when a device is unplugged, the thread currently
		//holding the lock will get an IO error (because the device
		//has disconnected) and return the lock fairly promptly.
		down(&IRIVER_S(sb)->lock);
		printk("[driver_disconnect] got lock.\n");
		IRIVER_S(sb)->active = 0;

		//The device has suddenly been unplugged.
		//IRIVER_S(sb)->dev = NULL;

		i = ifp_finalize(IRIVER_S(sb)->dev);
		if (i) {
			printk("[driver_disconnect] ifp_finalized returned %d.\n", i);
		}
		kfree(IRIVER_S(sb)->dev);
		IRIVER_S(sb)->dev = NULL;
		
		ifp_os_release_device(device_handle);
		device_handle = NULL;

		up(&IRIVER_S(sb)->lock);
	} else {
		//Normal unmount
		//printk("[driver_disconnect] shutting down normally.\n");
	}

	//printk("[driver_disconnect] closing usb interface..\n");

	return;
}

struct usb_driver iriver_usb_driver = {
	.owner =        THIS_MODULE,
	.name =         "iRiverManagerFS",
        .probe =        driver_probe,
	.disconnect =   driver_disconnect,
};

void iriver_put_super(struct super_block * sb) {
	struct iriver_sb_info * sbi;
	sbi = (struct iriver_sb_info *)sb->s_fs_info;

	if (sbi->active) {
		sbi->active = 0;
		cleanup(sb);
	}
	
	//printk("[iriver_put_super] start\n");
	if (IRIVER_S(sb)->dev == NULL) {
		printk("[iriver_put_super] device disconnected.. skipping dev\n");
	} else {
		put_ifp_device(IRIVER_S(sb)->dev, 0);
		IRIVER_S(sb)->dev = NULL;
	}

	kfree(sbi);
	sbi = sb->s_fs_info = NULL;
}

static void iriver_read_inode(struct inode *i) {
	printk("iriver_read_inode: start\n");
	printk("iriver_read_inode is ");
	if (i) {
		printk("%d\n", (int)i->i_ino);
	} else {
		printk("null\n");
	}
	return ;
}

static kmem_cache_t * iriver_inode_cachep;
kmem_cache_t * iriver_buffer_cachep;

static struct inode * iriver_alloc_inode(struct super_block *sb) {
	struct iriver_inode_info *ii = NULL;

	//printk("In iriver_alloc_inode\n");
	ii = (struct iriver_inode_info *)kmem_cache_alloc(iriver_inode_cachep,
		SLAB_KERNEL);
	if (ii == NULL) {
		return NULL;
	}
	ii->loaded = 0;
	ii->state = IFP_I_NORMAL;
	return &ii->vfs_inode;
}
static void iriver_destroy_inode(struct inode * inode) {
	kmem_cache_free(iriver_inode_cachep, IRIVER_I(inode));
	return ;
}

static struct super_operations iriver_ops = {
	.alloc_inode    = iriver_alloc_inode,
	.destroy_inode  = iriver_destroy_inode,
	.read_inode     = iriver_read_inode,
	.statfs		= iriver_statfs,
	.drop_inode	= generic_delete_inode,
	.put_super	= iriver_put_super,
};

//---Device detection and initialization.---//

//sleep in 1/10 of a second
static int geoff_sleep(int ds) {
	int i = 0;
	int timeout = ds*HZ/10;
	current->state = TASK_INTERRUPTIBLE;
	timeout = schedule_timeout(timeout);
	if (timeout) {
		i = -EINTR;
		//return i;
	}
	return i;
}

/*
 * Sanity check.  If the device is absent or completely FUBAR, simple actions
 * such as getting the model string, verison number and battery will likely
 * fail.
 *
 * While we're at it, let's print out that information so users have something
 * to email when they run into trouble.
 */
static int make_device_string(struct ifp_device * dev)
{
	int i;
	char * s = get_iriver_buffer();
	if (s == NULL) {
		return -ENOMEM;
	}

	i = ifp_device_info(dev, s, IFP_BUFFER_SIZE);
	if (i) {
		printk("[make_device_string] device info failed. err %d\n", i);
		goto out_0;
	}
	printk("iRiverFS: %s\n", s);

out_0:
	put_iriver_buffer(s);
	s = NULL;

	return i;
}

static int init_ifp_device(struct ifp_device * idev, struct usb_device * udev,
	void * context_for_interface) 
{
	int i = 0;

	if (udev == NULL || udev->actconfig == NULL
		|| udev->actconfig->interface == NULL
		|| udev->actconfig->interface[0] == NULL)
	{
		printk("iRiverFS: sorry, I couldn't claim the device's interface.  (Another driver might be using it.)\n");
		
		return -ENXIO;
	}

	//FIXME: find a more appropriate lock for this section.
	//from usb/core.c: "Callers must own the driver model's usb bus writelock."
	lock_kernel();
	i = usb_driver_claim_interface(&iriver_usb_driver,
		udev->actconfig->interface[0], context_for_interface);
	unlock_kernel();
	if (i) {
		printk("iRiverFS: sorry, I couldn't claim the device's interface.  (Another driver might be using it.)\n");
		i = -EBUSY;
		goto out_err0;
	}

	i = ifp_init(idev, udev);
	if (i) {
		printk("[get_ifp_device] initialization failed.. err %d\n",i);
		goto out_err1;
	}
	return i;

out_err1:
	lock_kernel();
	usb_driver_release_interface(&iriver_usb_driver,
		udev->actconfig->interface[0]);
	unlock_kernel();
out_err0:
	return i;

}

static int get_ifp_device(struct ifp_device ** pdev, void * context_for_interface) 
{
	struct ifp_device * dev;
	struct usb_device * device_handle;
	int i;

	dev = kmalloc(sizeof(struct ifp_device), GFP_KERNEL);
	*pdev = dev;
	if (!dev) {
		i = -ENOMEM;
		goto out_0;
	}

	device_handle = ifp_os_find_device();
	if (!device_handle) {
		printk("iRiverFS: sorry, I couldn't find a compatible iFP device.\n");
		i = -ENODEV;
		goto out_1;
	}

	i = init_ifp_device(dev, device_handle, context_for_interface);

	//don't attempt resets if the device was busy
	if (i == -ENXIO || i == -EBUSY) {
		goto out_2;
	}
	if (i) {
		printk("iRiverFS: device not responding, attempting software reset.\n");
		i = usb_reset_device(device_handle);
		if (i) { printk("error resetting device, err %d\n",i);}
		geoff_sleep(3);
		i = init_ifp_device(dev, device_handle, context_for_interface);
		if (i == 0) {
			printk("SUCCESSFUL RESET!!!! [first try] please tell geoff about this!\n");
		}
	}
	if (i) {
		printk("iRiverFS: device still not responding, attempting another reset.\n");
		i = usb_reset_device(device_handle);
		if (i) { printk("error resetting device, err %d\n",i);}
		geoff_sleep(12);
		i = init_ifp_device(dev, device_handle, context_for_interface);
		if (i == 0) {
			printk("SUCCESSFUL RESET!!!! [second try] please tell geoff about this!\n");
		}
	}
	if (i) {
		printk("iRiverFS: device not responding, giving up.\n");
		goto out_2;
	}

	return 0;

out_2:
	ifp_os_release_device(device_handle);
	//usb_put_dev(device_handle);
out_1:
	kfree(dev);
	*pdev = dev = NULL;
out_0:
	return i;
}

static int put_ifp_device(struct ifp_device * dev, int disconnected)
{
	int i = 0;
	struct usb_device * device_handle = dev->private;

	i = ifp_finalize(dev);
	if (i) {
		printk("[put_ifp_device] got %d from ifp_finalize.\n", i);
	}
	kfree(dev);
	dev = NULL;

	if (!disconnected) {
		lock_kernel();
		usb_driver_release_interface(&iriver_usb_driver,
			device_handle->actconfig->interface[0]);
		unlock_kernel();
	}
	
	ifp_os_release_device(device_handle);
	//usb_put_dev(device_handle);

	return 0;
}

enum {
	Opt_uid, Opt_gid,
	Opt_dmask, Opt_fmask,
	Opt_err,
};

static match_table_t  iriver_tokens = {
	{Opt_uid,   "uid=%u"},
	{Opt_gid,   "gid=%u"},
	{Opt_dmask, "dmask=%o"},
	{Opt_fmask, "fmask=%o"},
	{Opt_err,   NULL}
};

static int parse_options(char *options, struct iriver_mount_options * opts)
{
	int i = 0;

	char *p;
	substring_t args[MAX_OPT_ARGS];
	int option;

	if (!options) {
		return 0;
	}

	while ((p = strsep(&options, ",")) != NULL) {
		int token;
		if (!*p)
			continue;

		token = match_token(p, iriver_tokens, args);
		switch (token) {
		case Opt_uid:
			if (match_int(&args[0], &option))
				return 0;
			opts->uid = option;
			break;
		case Opt_gid:
			if (match_int(&args[0], &option))
				return 0;
			opts->gid = option;
			break;
		case Opt_dmask:
			if (match_octal(&args[0], &option))
				return 0;
			opts->dmask = option;
			break;
		case Opt_fmask:
			if (match_octal(&args[0], &option))
				return 0;
			opts->fmask = option;
			break;
		default:
			printk(KERN_ERR "unrecognized mount option %s or missing value\n", p);
			return -EINVAL;
		}
	}


	return i;
}

static int iriver_fill_super(struct super_block * sb, void * data, int silent)
{
	struct inode * inode;
	struct dentry * root;
	struct ifp_device * dev;
	int i = 0;

	sb->s_magic = IRIVERFS_MAGIC;
	sb->s_maxbytes = MAX_LFS_FILESIZE;
	sb->s_blocksize = IFP_BLOCK_SIZE_FAKE;
	sb->s_blocksize_bits = IFP_BLOCK_BITS_FAKE;
	sb->s_fs_info = kmalloc(sizeof(struct iriver_sb_info), GFP_KERNEL);
	if (!sb->s_fs_info) {
		i = -ENOMEM;
		goto out_0;
	}
	sema_init(&IRIVER_S(sb)->lock, 0);
	IRIVER_S(sb)->active        = 0;
	IRIVER_S(sb)->options.uid   = current->fsuid;
	IRIVER_S(sb)->options.gid   = current->fsgid;

	//if the user wants executable files, (s)he can always set the umask
	//manually.
	IRIVER_S(sb)->options.fmask = 0111 | current->fs->umask;
	IRIVER_S(sb)->options.dmask = 0000 | current->fs->umask;

	i = parse_options(data, &IRIVER_S(sb)->options);
	if (i) {
		goto out_1;
	}

	i = get_ifp_device(&(IRIVER_S(sb)->dev), sb);
	dev = IRIVER_S(sb)->dev;
	if (i) {
		if (i != -EBUSY) {
		    printk("[iriver_fill_super] error allocating device. i=%d\n",i);
		    if (i != -ENODEV) {
			printk("iRiverFS: the device isn't responding.  "
				"Try jiggling the handle.\n");
			//set the error code to EBUSY.. it makes more sense than
			//"fs not supported by kernel"
			i = -EBUSY;
		    }
		}
		goto out_1;
	}
	IRIVER_S(sb)->active = 1;
	IRIVER_S(sb)->cleanup = NULL;

	i = make_device_string(dev);
	if (i) {
		printk("[iriver_fill_super] error assembling the device string.\n");
		goto out_2;
	}

	sb->s_op = &iriver_ops;
	inode = iriver_get_inode(sb, S_IFDIR
		| (S_IRWXUGO & ~IRIVER_S(sb)->options.dmask), 0);
	if (!inode) {
		i = -ENOMEM;
		goto out_2;
	}
	IRIVER_I(inode)->loaded = 0;
	IRIVER_I(inode)->state = IFP_I_NORMAL;

	root = d_alloc_root(inode);
	if (!root) {
		i = -ENOMEM;
		goto out_3;
	}
	sb->s_root = root;

	//preloading inode/dentry info here would be too costly if someone put
	//this in their fstab.  What would be better is if we could fork a
	//child thread to do this in the background.

	up(&IRIVER_S(sb)->lock);
	return 0;

//	d_delete(root);
//	sb->s_root = root = NULL;
out_3:
	iput(inode);
	inode = NULL;
out_2:
	IRIVER_S(sb)->active = 0;
	put_ifp_device(IRIVER_S(sb)->dev, 0);
	IRIVER_S(sb)->dev = NULL;
out_1:
	//spin_unlock(&IRIVER_S(sb)->lock);
	up(&IRIVER_S(sb)->lock);
	kfree(sb->s_fs_info);
	sb->s_fs_info = NULL;
out_0:
	return i < 0 ? i : -EINVAL;
}

static struct super_block *iriver_get_sb(struct file_system_type *fs_type,
	int flags, const char *dev_name, void *data)
{
	return get_sb_nodev(fs_type, flags, data, iriver_fill_super);
}

static struct file_system_type iriver_fs_type = {
	.name		= "iriver",
	.get_sb		= iriver_get_sb,
	.kill_sb	= kill_litter_super,
};

static void init_once(void * foo, kmem_cache_t * cachep, unsigned long flags)
{
	struct iriver_inode_info *ei = (struct iriver_inode_info *) foo;

	//printk("[init_once] %p\n", foo);

	if ((flags & (SLAB_CTOR_VERIFY|SLAB_CTOR_CONSTRUCTOR)) ==
	    SLAB_CTOR_CONSTRUCTOR)
	inode_init_once(&ei->vfs_inode);
}
static void destroy_twice(void * foo, kmem_cache_t * cachep, unsigned long flags)
{
	//printk("[destroy_twice] %p\n", foo);
}
static int init_inode_cache(void)
{
	iriver_inode_cachep = kmem_cache_create("iriver_inode_cachep",
		sizeof(struct iriver_inode_info), 0,
		SLAB_HWCACHE_ALIGN|SLAB_RECLAIM_ACCOUNT, init_once, destroy_twice);

	iriver_buffer_cachep = kmem_cache_create("iriver_buffer_cachep",
		0x0400, 0,
		SLAB_HWCACHE_ALIGN|SLAB_RECLAIM_ACCOUNT, NULL, NULL);

	if (iriver_inode_cachep == NULL)
		return -ENOMEM;
	return 0;
}
static void destroy_inode_cache(void)
{
	if (kmem_cache_destroy(iriver_inode_cachep)) {
		printk(KERN_INFO "iriver_inode_cache: not all structures were freed\n");
	}
	iriver_inode_cachep = NULL;

	if (kmem_cache_destroy(iriver_buffer_cachep)) {
		printk(KERN_INFO "iriver_buffer_cache: not all structures were freed\n");
	}
	iriver_buffer_cachep = NULL;
}

static int __init init_iriver_fs(void)
{
	int i;

	i = init_inode_cache();
	if (i) {
		printk("gfs.init_iriver_fs: error creating inode cache.\n");
		goto out1;
	}

	i = usb_register(&iriver_usb_driver);
	if (i) {
		printk("gfs.init_iriver_fs: error registering usb driver. (%d)\n",i);
		goto out_cache;
	}

	i = register_filesystem(&iriver_fs_type);
	if (i)  {
		printk("gfs.init_iriver_fs: error registering file system.\n");
		goto out_usb;
	}
	printk("iRiverFS: successfully registered and initialized.\n");

	return 0;

out_usb:
	printk("[init_iriver_fs] error, cleaning up usb.\n");
	usb_deregister(&iriver_usb_driver);
out_cache:
	printk("[init_iriver_fs] error, cleaning up cache.\n");
	destroy_inode_cache();
out1:
	printk("[init_iriver_fs] error somewhere.\n");
	return i;
}

static void __exit exit_iriver_fs(void)
{
	unregister_filesystem(&iriver_fs_type);
	usb_deregister(&iriver_usb_driver);
	destroy_inode_cache();
}

module_init(init_iriver_fs)
module_exit(exit_iriver_fs)

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("iRiver iFP Manager, linux vfs edition");
MODULE_AUTHOR("Geoff Oakham");

